/*
 * Copyright (c) 2014 Cloudera, Inc. All rights reserved.
 */

package com.cloudera.keytrustee;

import com.cloudera.keytrustee.impl.ClientFactoryImpl;
import com.cloudera.keytrustee.impl.FileTokenStore;
import com.cloudera.keytrustee.util.Environment;
import com.google.common.base.Preconditions;
import org.apache.commons.pool2.BasePooledObjectFactory;
import org.apache.commons.pool2.ObjectPool;
import org.apache.commons.pool2.PooledObject;
import org.apache.commons.pool2.impl.AbandonedConfig;
import org.apache.commons.pool2.impl.DefaultPooledObject;
import org.apache.commons.pool2.impl.GenericObjectPool;
import org.apache.commons.pool2.impl.GenericObjectPoolConfig;
import org.apache.hadoop.conf.Configuration;

import java.io.File;
import java.io.IOException;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

/**
 * Factory for obtaining a KeyTrustee ClientConnection. Implements ClientConnection
 * pooling using org.apache.commons.pool2.BasePooledObjectFactory.
 */
class PooledClientConnectionFactory extends BasePooledObjectFactory<ClientConnection> {

  public static final int DEFAULT_MAX_IDLE = 5;
  public static final int DEFAULT_MAX_SIZE = 10;
  public static final int DEFAULT_MIN_IDLE = 2;
  public static final int DEFAULT_ABANDONED_TIMEOUT = 180;

  private final ClientFactory keyTrusteeClientFactory;
  private final File keyTrusteeConfDir;

  private final Lock readLock;
  private final Lock writeLock;

  private final ObjectPool<ClientConnection> connectionPool;
  private final TrusteeClientConnectionPool keyTrusteeConnectionPool;

  private GenericObjectPoolConfig objectPoolConfig;
  private AbandonedConfig abandonedConfig;

  public PooledClientConnectionFactory(File keyTrusteeConfDir, TrusteeKeyProviderConfiguration tkpConfig) throws KeyTrusteeException, IOException {
    Preconditions.checkNotNull(keyTrusteeConfDir, "Null keyTrusteeConfDir passed to constructor.");

    this.keyTrusteeClientFactory = new ClientFactoryImpl();
    this.keyTrusteeConfDir = keyTrusteeConfDir;

    initializeConnectionPoolConfig(tkpConfig);
    this.connectionPool = new GenericObjectPool<ClientConnection>(this, this.objectPoolConfig, this.abandonedConfig);
    this.keyTrusteeConnectionPool = new TrusteeClientConnectionPool(this.connectionPool);

    ReadWriteLock lock = new ReentrantReadWriteLock(true);
    readLock = lock.readLock();
    writeLock = lock.writeLock();
  }

  private void initializeConnectionPoolConfig(TrusteeKeyProviderConfiguration tkpConfig) {
    this.objectPoolConfig = new GenericObjectPoolConfig();

    objectPoolConfig.setMaxIdle(tkpConfig.getMaxIdle());
    objectPoolConfig.setMaxTotal(tkpConfig.getMaxTotal());
    objectPoolConfig.setMinIdle(tkpConfig.getMinIdle());
    this.abandonedConfig = new AbandonedConfig();
    abandonedConfig.setRemoveAbandonedTimeout(tkpConfig.getAbandonedTimeout()); //seconds
    abandonedConfig.setRemoveAbandonedOnMaintenance(true);
  }

  @Override
  public ClientConnection create() throws Exception {
    try {
      writeLock.lock();
      ClientConnection keyTrusteeConnection = null;
      ClientInfo keyTrusteeClient = keyTrusteeClientFactory.openClient(keyTrusteeConfDir);
      ServerInfo keyTrusteeServer = keyTrusteeClient.getServerInfo();
      if (keyTrusteeServer.isTokenSync()) {
        FileTokenStore tokenStore = new FileTokenStore(keyTrusteeConfDir, keyTrusteeServer);
        keyTrusteeConnection = keyTrusteeClientFactory.connect(keyTrusteeClient, keyTrusteeServer, tokenStore);
      } else {
        keyTrusteeConnection = keyTrusteeClientFactory.connect(keyTrusteeClient, keyTrusteeServer);
      }
      return keyTrusteeConnection;
    } finally {
      writeLock.unlock();
    }
  }

  @Override
  public PooledObject<ClientConnection> wrap(ClientConnection clientConnection) {
    Preconditions.checkNotNull(clientConnection);
    return new DefaultPooledObject<ClientConnection>(clientConnection);
  }

  TrusteeClientConnectionPool getKeyTrusteeConnectionPool() {
    return this.keyTrusteeConnectionPool;
  }

}
